# -*- coding: utf-8 -*-

"""
Config module for parsing input args for lbsntransform package.
"""

# pylint: disable=no-member

from pathlib import Path
import argparse
import os
import sys
from shapely import geos
import logging
from lbsnstructure.lbsnstructure_pb2 import lbsnPost


class BaseConfig():
    def __init__(self):
        """Set Default Config options here

        - or define options as input args
        """
        self.origin = 3
        self.is_local_input = False
        self.local_file_type = 'json'
        self.input_path = None
        self.is_stacked_json = False
        self.dbuser_input = 'example-user-name'
        self.dbpassword_input = 'example-user-password'
        self.dbserveraddress_input = '222.22.222.22'
        self.dbserverport_input = 5432
        self.dbname_input = 'test_db2'
        self.dbuser_output = None
        self.dbpassword_output = None
        self.dbserveraddress_output = None
        self.dbserverport_output = 5432
        self.dbname_output = None
        self.transferlimit = None
        self.transfer_count = 50000
        self.number_of_records_to_fetch = 10000
        self.transfer_reactions = True
        self.disable_reactionpost_ref = False
        self.ignore_non_geotagged = False
        self.startwith_db_rownumber = 0
        self.endwith_db_rownumber = None
        self.debug_mode = 'INFO'
        self.geocode_locations = False
        self.ignore_input_source_list = False
        self.input_lbsn_type = None
        self.map_relations = False
        self.csv_output = False
        self.csv_suppress_linebreaks = True
        self.csv_delim = None
        self.recursive_load = False
        self.skip_until_file = None
        self.min_geoaccuracy = None
        self.logging_level = logging.INFO
        self.source_web = False

        BaseConfig.set_options()

    def parseArgs(self):
        """Process input *args

        All args are optional, but some groups need to be defined together.
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('-sO', "--Origin",
                            default=self.origin,
                            help='Type of input source. '
                            'Defaults to 3: Twitter '
                            '(1 - Instagram, 2 - Flickr, 3 - Twitter)')
        # Local Input
        local_input_args = parser.add_argument_group('Local Input')
        local_input_args.add_argument('-lI', "--LocalInput",
                                      action='store_true',
                                      default=False,
                                      help='Process local json or csv')
        local_input_args.add_argument('-lT', "--LocalFileType",
                                      default=self.local_file_type,
                                      help='If localread, '
                                      'specify filetype (json, csv etc.)')
        local_input_args.add_argument('-iP', "--InputPath",
                                      default=self.input_path,
                                      help='Optionally provide path to '
                                      'input folder, otherwise '
                                      './Input/ will be used. You can also '
                                      'provide a web-url starting with http')
        local_input_args.add_argument('-iS', "--isStackedJson",
                                      action='store_true',
                                      default=False,
                                      help='Typical form is '
                                      '[{json1},{json2}], '
                                      'if is_stacked_json is True: '
                                      'will process stacked jsons in the form '
                                      'of {json1}{json2} (no comma)')
        # DB Output
        dboutput_args = parser.add_argument_group('DB Output')
        dboutput_args.add_argument('-pO', "--dbPassword_Output",
                                   default=self.dbpassword_output,
                                   help='')
        dboutput_args.add_argument('-uO', "--dbUser_Output",
                                   default=self.dbuser_output,
                                   help='Default: example-user-name2')
        dboutput_args.add_argument('-aO', "--dbServeraddressOutput",
                                   default=self.dbserveraddress_output,
                                   help='e.g. 111.11.11.11 . Optionally add '
                                   'port to use, e.g. 111.11.11.11:5432. '
                                   '5432 is the default port')
        dboutput_args.add_argument('-nO', "--dbNameOutput",
                                   default=self.dbname_output,
                                   help='e.g.: test_db')
        # DB Input
        dbinput_args = parser.add_argument_group('DB Input')
        dbinput_args.add_argument('-pI', "--dbPassword_Input",
                                  default=self.dbpassword_input,
                                  help='')
        dbinput_args.add_argument('-uI', "--dbUser_Input",
                                  default=self.dbuser_input,
                                  help='Default: example-user-name')
        dbinput_args.add_argument('-aI', "--dbServeraddressInput",
                                  default=self.dbserveraddress_input,
                                  help='e.g. 111.11.11.11. Optionally add '
                                  'port to use, e.g. 111.11.11.11:5432. '
                                  '5432 is the default port')
        dbinput_args.add_argument('-nI', "--dbNameInput",
                                  default=self.dbname_input,
                                  help='e.g.: test_db')
        # Additional args
        settings_args = parser.add_argument_group('Additional settings')
        settings_args.add_argument('-t', "--transferlimit",
                                   default=self.transferlimit,
                                   help='')
        settings_args.add_argument('-tC', "--transferCount",
                                   default=self.transfer_count,
                                   help='Default to 50k: After how many '
                                   'parsed records should the result be '
                                   'transferred to the DB. Larger values '
                                   'improve speed, because duplicate '
                                   'check happens in Python and not in '
                                   'Postgres Coalesce; larger values are '
                                   'heavier on memory.')
        settings_args.add_argument('-nR', "--numberOfRecordsToFetch",
                                   default=self.number_of_records_to_fetch,
                                   help='')
        settings_args.add_argument(
            '-tR', "--disableTransferReactions",
            action='store_true',
            help='')
        settings_args.add_argument(
            '-rR', "--disableReactionPostReferencing",
            action='store_true',
            default=False,
            help='Enable this option in args to prevent empty posts stored '
            'due to Foreign Key Exists Requirement '
            '0 = Save Original Tweets of Retweets in "posts"; 1 = do not '
            'store Original Tweets of Retweets; !Not implemented: 2 = Store '
            'Original Tweets of Retweets as "post_reactions"')
        settings_args.add_argument('-iG', "--ignoreNonGeotagged",
                                   action='store_true',
                                   default=self.ignore_non_geotagged,
                                   help='')
        settings_args.add_argument('-rS', "--startWithDBRowNumber",
                                   default=self.startwith_db_rownumber,
                                   help='')
        settings_args.add_argument('-rE', "--endWithDBRowNumber",
                                   default=self.endwith_db_rownumber,
                                   help='')
        settings_args.add_argument('-d', "--debugMode",
                                   default=self.debug_mode,
                                   help='Needs to be implemented.')
        settings_args.add_argument('-gL', "--geocodeLocations",
                                   default=self.geocode_locations,
                                   help='Defaults to None. '
                                   'Provide path to CSV file with '
                                   'location geocodes (CSV Structure: '
                                   'lat, lng, name)')
        settings_args.add_argument('-igS', "--ignoreInputSourceList",
                                   default=self.ignore_input_source_list,
                                   help='Provide a list of input_source '
                                   'types that will be ignored (e.g. to '
                                   'ignore certain bots etc.)')
        settings_args.add_argument('-iT', "--inputType",
                                   default=self.input_lbsn_type,
                                   help='Input type, e.g. "post", "profile", '
                                   '"friendslist", "followerslist" etc.')
        settings_args.add_argument('-mR', "--mapFullRelations",
                                   action='store_true',
                                   help='Defaults to False. Set to true '
                                   'to map full relations, e.g. many-to-many '
                                   'relationships such as user_follows, '
                                   'user_friend, user_mentions etc. are '
                                   'mapped in a separate table')
        settings_args.add_argument('-CSV', "--CSVOutput",
                                   action='store_true',
                                   default=self.csv_output,
                                   help='Set to True to Output all '
                                   'Submit values to CSV')
        settings_args.add_argument('-CSVal', "--CSVallowLinebreaks",
                                   action='store_true', default=False,
                                   help='If set to False will not '
                                   'remove intext-linebreaks (\r or \n) '
                                   'in output CSVs')
        settings_args.add_argument('-CSVdelim', "--CSVdelimiter",
                                   default=None,
                                   help=repr(
                                       'Provide CSV delimiter to use. '
                                       'Default is comma(,). Note: to pass tab, '
                                       'use variable substitution ($"\t")'))
        settings_args.add_argument('-rL', "--recursiveLoad",
                                   action='store_true', default=False,
                                   help='Process Input Directories '
                                   'recursively (depth: 2)')
        settings_args.add_argument('-sF', "--skipUntilFile",
                                   default=self.skip_until_file,
                                   help='If local input, skip all files '
                                   'until file with name x appears '
                                   '(default: start immediately)')
        settings_args.add_argument('-mGA', "--minGeoAccuracy",
                                   default=self.min_geoaccuracy,
                                   help='Set to "latlng", "place", '
                                   'or "city" to limit output based '
                                   'on min geoaccuracy')

        args = parser.parse_args()
        if args.LocalInput:
            self.is_local_input = True
            self.local_file_type = args.LocalFileType
            if args.isStackedJson:
                self.is_stacked_json = True
            if not args.InputPath:
                self.input_path = Path.cwd() / "01_Input"
                print(f'Using Path: {self.input_path}')
            else:
                if str(args.InputPath).startswith('http'):
                    self.input_path = str(args.InputPath)
                    self.source_web = True
                else:
                    input_path = Path(args.InputPath)
                    self.input_path = input_path
        else:
            self.dbuser_input = args.dbUser_Input
            self.dbpassword_input = args.dbPassword_Input
            input_ip, input_port = BaseConfig.get_ip_port(
                args.dbServeraddressInput)
            self.dbserveraddress_input = input_ip
            if input_port:
                self.dbserverport_input = input_port
            self.dbname_input = args.dbNameInput
        if args.Origin:
            self.origin = int(args.Origin)
        if args.geocodeLocations:
            self.geocode_locations = f'{os.getcwd()}\\'
            f'{args.geocodeLocations}'
        if args.ignoreInputSourceList:
            self.ignore_input_source_list = f'{os.getcwd()}\\'
            f'{args.ignoreInputSourceList}'
        if args.dbUser_Output:
            self.dbuser_output = args.dbUser_Output
            self.dbpassword_output = args.dbPassword_Output
            output_ip, output_port = BaseConfig.get_ip_port(
                args.dbServeraddressOutput)
            self.dbserveraddress_output = output_ip
            if output_port:
                self.dbserverport_output = output_port
            self.dbname_output = args.dbNameOutput
        if args.transferlimit:
            self.transferlimit = int(args.transferlimit)
            if self.transferlimit == 0:
                self.transferlimit = None
        if args.transferCount:
            self.transfer_count = int(args.transferCount)
        if args.numberOfRecordsToFetch:
            self.number_of_records_to_fetch = int(args.numberOfRecordsToFetch)
        if args.disableTransferReactions is True:
            self.transfer_reactions = False
        if args.disableReactionPostReferencing:
            self.disable_reactionpost_ref = True
        self.ignore_non_geotagged = args.ignoreNonGeotagged
        if args.startWithDBRowNumber:
            self.startwith_db_rownumber = int(args.startWithDBRowNumber)
        if args.endWithDBRowNumber:
            self.endwith_db_rownumber = int(args.endWithDBRowNumber)
        self.debug_mode = args.debugMode
        if args.inputType:
            self.input_lbsn_type = args.inputType
        if args.mapFullRelations:
            self.map_relations = True
        if args.CSVOutput:
            self.csv_output = True
        if args.CSVallowLinebreaks:
            self.csv_suppress_linebreaks = False
        if args.CSVdelimiter:
            self.csv_delim = args.CSVdelimiter
        if args.recursiveLoad:
            self.recursive_load = True
        if args.skipUntilFile:
            self.skip_until_file = args.skipUntilFile
        if args.minGeoAccuracy:
            self.min_geoaccuracy = self.check_geoaccuracy_input(
                args.minGeoAccuracy)

    @staticmethod
    def check_geoaccuracy_input(geoaccuracy_string):
        """Checks geoaccuracy input string and matches
        against proto buf spec
        """
        if geoaccuracy_string == 'latlng':
            return lbsnPost.LATLNG
        elif geoaccuracy_string == 'place':
            return lbsnPost.PLACE
        elif geoaccuracy_string == 'city':
            return lbsnPost.CITY
        else:
            print("Unknown geoaccuracy.")
            return None

    @staticmethod
    def get_ip_port(ip_string: str):
        """Gets IP and port, if available

        Arguments:
            ip_string {str} -- String containing IPV4 and
            possibly port attached with : character
        """
        port = None
        ip = None
        ip_string_spl = ip_string.split(":")
        if len(ip_string_spl) == 2:
            ip = ip_string_spl[0]
            port = ip_string_spl[1]
        else:
            ip = ip_string
        return ip, port

    @staticmethod
    def set_options():
        """Includes global options in other packages to be set
        prior execution"""
        # tell shapely to include the srid when generating WKBs
        geos.WKBWriter.defaults['include_srid'] = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
lbsntransform package script to load, format and store data
from and to common lbsn structure

Import options:
    - Postgres Database or
    - local CSV/json/stacked json import
Output options:
    - Postgres Database or
    - local ProtoBuf and CSV Import (prepared for Postgres /Copy)
"""

__author__ = "Alexander Dunkel"
__license__ = "GNU GPLv3"
# version: see version.py

import io
import logging
import sys

from lbsntransform.classes.helper_functions import HelperFunctions as HF
from lbsntransform.classes.helper_functions import TimeMonitor
from lbsntransform.classes.load_data import LoadData
from lbsntransform.classes.submit_data import LBSNTransfer
from lbsntransform.config.config import BaseConfig
from lbsntransform.lbsntransform_ import LBSNTransform


def main():
    """ Main function to process data from postgres db or local file input
        to postgres db or local file output
    """

    # Load Config, will be overwritten if args are given
    config = BaseConfig()
    # Parse args
    config.parseArgs()

    # initialize lbsntransform
    lbsntransform = LBSNTransform(
        origin_id=config.origin,
        logging_level=config.logging_level,
        is_local_input=config.is_local_input,
        csv_output=config.csv_output,
        csv_suppress_linebreaks=config.csv_suppress_linebreaks,
        dbuser_output=config.dbuser_output,
        dbserveraddress_output=config.dbserveraddress_output,
        dbname_output=config.dbname_output,
        dbpassword_output=config.dbpassword_output,
        dbserverport_output=config.dbserverport_output,
        dbuser_input=config.dbuser_input,
        dbserveraddress_input=config.dbserveraddress_input,
        dbname_input=config.dbname_input,
        dbpassword_input=config.dbpassword_input,
        dbserverport_input=config.dbserverport_input)

    # initialize converter class
    # depending on lbsn origin
    # e.g. 1 = Instagram,
    #      2 = Flickr, 2.1 = Flickr YFCC100m,
    #      3 = Twitter)
    importer = HF.load_importer_mapping_module(
        config.origin)

    # initialize input reader
    input_data = LoadData(
        importer=importer,
        is_local_input=config.is_local_input,
        startwith_db_rownumber=config.startwith_db_rownumber,
        skip_until_file=config.skip_until_file,
        cursor_input=lbsntransform.cursor_input,
        input_path=config.input_path,
        recursive_load=config.recursive_load,
        local_file_type=config.local_file_type,
        endwith_db_rownumber=config.endwith_db_rownumber,
        is_stacked_json=config.is_stacked_json,
        csv_delim=config.csv_delim,
        input_lbsn_type=config.input_lbsn_type,
        geocode_locations=config.geocode_locations,
        ignore_input_source_list=config.ignore_input_source_list,
        disable_reactionpost_ref=config.disable_reactionpost_ref,
        map_relations=config.map_relations,
        transfer_reactions=config.transfer_reactions,
        ignore_non_geotagged=config.ignore_non_geotagged,
        min_geoaccuracy=config.min_geoaccuracy,
        source_web=config.source_web)

    # Manually add entries that need submission prior to parsing data
    # add_bundestag_group_example(import_mapper)

    # init time monitoring
    how_long = TimeMonitor()

    # read and process unfiltered input records from csv
    # start settings
    with input_data as records:
        for record in records:
            lbsntransform.add_processed_records(record)
            # report progress
            if lbsntransform.processed_total % 1000 == 0:
                print(
                    f'{lbsntransform.processed_total} '
                    f'input records processed (up to '
                    f'{input_data.continue_number}). '
                    f'Skipped '
                    f'{input_data.import_mapper.skipped_low_geoaccuracy} '
                    f'due to low geoaccuracy. '
                    f'Count per type: '
                    f'{lbsntransform.lbsn_records.get_type_counts()}'
                    f'records.', end='\r')

    # finalize output (close db connection, submit remaining)
    print(f'Transferring remaining '
          f'{lbsntransform.lbsn_records.count_glob} to db.. '
          f'{HF.null_notice(input_data.import_mapper.null_island)})')
    lbsntransform.finalize_output()

    # final report
    print(f'\n\nProcessed {lbsntransform.processed_total} input records '
          f'(Input {input_data.start_number} to '
          f'{input_data.continue_number}). '
          f'Skipped {input_data.import_mapper.skipped_low_geoaccuracy} '
          f'due to low geoaccuracy.')

    input(f'Done. {how_long.stop_time()}')


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
Module for mapping Flickr YFCC100M dataset to common LBSN Structure.
"""

import csv
import logging
import re
import sys
from codecs import escape_decode
from decimal import Decimal
from urllib.parse import unquote

# for debugging only:
from google.protobuf import text_format
from lbsnstructure.lbsnstructure_pb2 import (CompositeKey, Language,
                                             RelationshipKey, lbsnCity,
                                             lbsnCountry, lbsnOrigin,
                                             lbsnPlace, lbsnPost,
                                             lbsnPostReaction,
                                             lbsnRelationship, lbsnUser,
                                             lbsnUserGroup)

from .helper_functions import HelperFunctions as HF
from .helper_functions import LBSNRecordDicts

# pylint: disable=no-member


class FieldMappingYFCC100M():
    """ Provides mapping function from Flickr endpoints to
        protobuf lbsnstructure
    """
    ORIGIN_NAME = "Flickr yfcc100m"
    ORIGIN_ID = 2

    def __init__(self,
                 disable_reaction_post_referencing=False,
                 geocodes=False,
                 map_full_relations=False,
                 map_reactions=True,
                 ignore_non_geotagged=False,
                 ignore_sources_set=set(),
                 min_geoaccuracy=None):
        # We're dealing with Flickr in this class, lets create the OriginID
        # globally
        # this OriginID is required for all CompositeKeys
        origin = lbsnOrigin()
        origin.origin_id = lbsnOrigin.FLICKR
        self.origin = origin
        self.null_island = 0
        self.log = logging.getLogger('__main__')  # get the main logger object
        self.skipped_count = 0
        self.skipped_low_geoaccuracy = 0
        # some records in YFCC100m are larger
        # than the default csv limit in python
        # of 131072
        csv.field_size_limit(500000)
        # self.disableReactionPostReferencing = disableReactionPostReferencing
        # self.mapFullRelations = mapFullRelations
        # self.geocodes = geocodes
        self.lic_dict = {
            "All Rights Reserved": 0,
            "Attribution-NonCommercial-ShareAlike License": 1,
            "Attribution-NonCommercial License": 2,
            "Attribution-NonCommercial-NoDerivs License": 3,
            "Attribution License": 4,
            "Attribution-ShareAlike License": 5,
            "Attribution-NoDerivs License": 6,
            "No known copyright restrictions": 7,
            "United States Government Work": 8,
            "Public Domain Dedication (CC0)": 9,
            "Public Domain Mark": 10
        }

    def parse_csv_record(self, record):
        """Entry point for flickr CSV data:
            - Decide if CSV record contains user-info or post-info
            - Skip empty or malformed records

        Attributes:
        record    A single row from CSV, stored as list type.
        """
        if len(record) == 2:
            # Flickr place record dirty detection
            lbsn_records = self.extract_flickr_place(
                record)
            return lbsn_records
        elif len(record) < 12 or not record[0].isdigit():
            # skip erroneous entries + log
            self.skipped_count += 1
            return
        else:
            # process regular Flickr yfcc100m record
            lbsn_records = self.extract_flickr_post(record)
            return lbsn_records

    def extract_flickr_post(self, record):
        """Main function for processing Flickr YFCC100M CSV entry.
           This mothod is adapted to a special structure, adapt if needed.

        To Do:
            - parameterize column numbers and structure
            - provide external config-file for specific CSV structures
            - currently not included in lbsn mapping are MachineTags,
              GeoContext (indoors, outdoors), WoeId
              and some extra attributes only present for Flickr

        Overview of available columns and examples:
        0 row-number    -   0
        1 unknown - 4e2f7a26a1dfbf165a7e30bdabf7e72a
        2 Photo/video identifier    -   6985418911
        3 User NSID     -   4e2f7a26a1dfbf165a7e30bdabf7e72a
        4 User nickname     -   39089491@N00
        5 Date taken    -   2012-02-16 09:56:37.0
        6 Date uploaded     -   1331840483
        7 Capture device    -   Canon+PowerShot+ELPH+310+HS
        8 Title     -   IMG_0520
        9 Description      -     My vacation
        10 User tags (comma-separated)   -   canon,canon+powershot+hs+310
        11 Machine tags (comma-separated)   - landscape, hills, water
        12 Longitude    -   -81.804885
        13 Latitude     -   24.550558
        14 Accuracy -   12
        15 Photo/video page URL -   http://www.flickr.com/photos/39089491@N00/6985418911/
        16 Photo/video download URL -   http://farm8.staticflickr.com/7205/6985418911_df7747990d.jpg
        17 License name -   Attribution-NonCommercial-NoDerivs License
        18 License URL  -   http://creativecommons.org/licenses/by-nc-nd/2.0/
        19 Photo/video server identifier    -   7205
        20 Photo/video farm identifier  -   8
        21 Photo/video secret   -   df7747990d
        22 Photo/video secret original  -   692d7e0a7f
        23 Extension of the original photo  -   jpg
        24 Marker (0 ¼ photo, 1 ¼ video)    -   0
        """
        # note that one input record may contain many lbsn records
        # therefore, return list of processed records
        lbsn_records = []
        # start mapping input to lbsn_records
        post_guid = record[1]
        if not HF.check_notice_empty_post_guid(post_guid):
            return None
        post_record = HF.new_lbsn_record_with_id(lbsnPost(),
                                                 post_guid,
                                                 self.origin)
        user_record = HF.new_lbsn_record_with_id(lbsnUser(),
                                                 record[3],
                                                 self.origin)
        user_record.user_name = unquote(record[4]).replace(
            '+', ' ')
        user_record.url = f'http://www.flickr.com/photos/{user_record.pkey.id}/'
        if user_record:
            post_record.user_pkey.CopyFrom(user_record.pkey)
        lbsn_records.append(user_record)
        post_record.post_latlng = self.flickr_extract_postlatlng(record)
        geoaccuracy = FieldMappingYFCC100M.flickr_map_geoaccuracy(
            record[14])
        if geoaccuracy:
            post_record.post_geoaccuracy = geoaccuracy
        # place record available in separate yfcc100m dataset
        # if record[1]:
            # we need some information from postRecord to create placeRecord
            # (e.g.  user language, geoaccuracy, post_latlng)
            # some of the information from place will also modify postRecord
            # place_record = HF.new_lbsn_record_with_id(lbsnPlace(),
            #                                           record[1],
            #                                           self.origin)
            # lbsn_records.append(place_record)
            # post_record.place_pkey.CopyFrom(place_record.pkey)
        post_record.post_publish_date.CopyFrom(
            HF.parse_timestamp_string_to_protobuf(record[6]))
        post_created_date = HF.parse_csv_datestring_to_protobuf(
            record[5], t_format='%Y-%m-%d %H:%M:%S.%f')
        if post_created_date:
            post_record.post_create_date.CopyFrom(
                post_created_date
            )
        post_record.post_views_count = 0
        post_record.post_comment_count = 0
        post_record.post_like_count = 0
        post_record.post_url = record[15]
        # YFCC100M dataset contains HTML codes (%20) and
        # space character is replaced by +
        post_record.post_body = unquote(record[9]).replace(
            '+', ' ')
        post_record.post_title = unquote(record[8]).replace(
            '+', ' ')
        post_record.post_thumbnail_url = record[16]  # note: fullsize url!
        # split tags by , and + because by lbsn-spec,
        # tags are limited to single word
        record_tags_list = list(
            set(filter(
                None,
                [unquote(tag)
                 for tag in re.split("[,+]+", record[10])]
            )))
        if record_tags_list:
            for tag in record_tags_list:
                tag = FieldMappingYFCC100M.clean_tags_from_flickr(tag)
                post_record.hashtags.append(tag)
        record_machine_tags = list(
            set(filter(None, [unquote(mtag) for mtag in re.split("[,+]+", record[11])])))
        if 'video' in record_machine_tags:
            # all videos appear to have 'video' in machine tags
            post_record.post_type = lbsnPost.VIDEO
        else:
            post_record.post_type = lbsnPost.IMAGE
        # replace text-string of content license by integer-id
        post_record.post_content_license = self.get_license_number_from_license_name(
            record[17])
        lbsn_records.append(post_record)
        return lbsn_records

    def extract_flickr_place(self, record):
        """Main function for processing Flickr YFCC100M Place CSV entry
        Place records are available from a separate yfcc100m dataset
        and are joined based on photo guid
        """
        # note that one input record may contain many lbsn records
        # therefore, return list of processed records
        lbsn_records = []
        # start mapping input to lbsn_records
        post_guid = record[0]
        if not record[1]:
            # skip empty records
            return None
        # place records in yfcc100m contain multiple entries, e.g.
        # 24703176:Admiralty:Suburb,24703128:Central+and+Western:Territory,24865698:Hong+Kong:Special Administrative Region,28350827:Asia%2FHong_Kong:Timezone
        place_records = record[1].split(",")
        for place_record in place_records:
            lbsn_place_record = \
                FieldMappingYFCC100M.process_place_record(
                    place_record, self.origin)
            lbsn_records.append(lbsn_place_record)
        # update post record with entries from place record
        post_record = HF.new_lbsn_record_with_id(
            lbsnPost(), post_guid, self.origin)
        for place_record in lbsn_records:
            if isinstance(place_record, lbsnCountry):
                post_record.country_pkey.CopyFrom(place_record.pkey)
            if isinstance(place_record, lbsnCity):
                post_record.city_pkey.CopyFrom(place_record.pkey)
            # either city or place, Twitter user cannot attach both (?)
            elif isinstance(place_record, lbsnPlace):
                post_record.place_pkey.CopyFrom(place_record.pkey)
        lbsn_records.append(post_record)
        return lbsn_records

    @staticmethod
    def process_place_record(place_record, origin):
        """Assignment of Flickr place types to lbsnstructure
        hierarchy: Country, City, Place
        Original Flickr place types, which are more detailed,
        are stored in sub_type field
        """
        FLICKR_COUNTRY_MATCH = set([
            "country", "timezone", "state", "territory",
            "state/territory", "land", "arrondissement", "prefecture",
            "island", "disputed territory", "overseas collectivity",
            "overseas region", "island group"
        ])
        FLICKR_CITY_MATCH = set([
            "city", "town", "region", "special administrative region",
            "county", "district", "zip", "province", "suburb",
            "district/county", "autonomous community", "municipality",
            "department", "district/town/township", "township",
            "historical town", "governorate", "commune", "canton",
            "muhafazah", "local government area", "division", "periphery",
            "zone", "sub-region", "district/town/county", "sub-district",
            "parish", "federal district", "neighborhood", "chome", "ward",
            "sub-division", "community", "autonomous region",
            "municipal region", "emirate", "autonomous province",
            "historical county", "constituency", "entity", "colloquial",
            "circle", "quarter", "dependency", "sub-prefecture"
        ])
        FLICKR_PLACE_MATCH = set([
            "poi", "land feature", "estate", "airport", "drainage",
            "miscellaneous"
        ])
        place_record_split = place_record.split(":")
        if not len(place_record_split) == 3:
            raise ValueError(
                f'Malformed place entry:\n'
                f'place_record: {place_record}')
        place_guid = unquote(place_record_split[0])
        place_name = unquote(place_record_split[1]).replace(
            '+', ' ')
        place_type = unquote(place_record_split[2])
        place_type_lw = place_type.lower()
        place_type_lw_split = place_type_lw.split("/")
        # assignment
        if any(ptls in FLICKR_COUNTRY_MATCH for ptls in place_type_lw_split):
            lbsn_place_record = HF.new_lbsn_record_with_id(
                lbsnCountry(), place_guid, origin)
        elif any(ptls in FLICKR_CITY_MATCH for ptls in place_type_lw_split):
            lbsn_place_record = HF.new_lbsn_record_with_id(
                lbsnCity(), place_guid, origin)
        elif any(ptls in FLICKR_PLACE_MATCH for ptls in place_type_lw_split):
            lbsn_place_record = HF.new_lbsn_record_with_id(
                lbsnPlace(), place_guid, origin)
        else:
            logging.getLogger('__main__').WARNING(
                f'Could not assign place type {place_type_lw}\n'
                f'found in place_record: {place_record}\n'
                f'Will assign default "Place"')
            lbsn_place_record = HF.new_lbsn_record_with_id(
                lbsnPlace(), place_guid, origin)
        lbsn_place_record.name = place_name
        if isinstance(lbsn_place_record, lbsnCity):
            # record sub types only for city and place
            lbsn_place_record.sub_type = place_type
        elif isinstance(lbsn_place_record, lbsnPlace):
            lbsn_place_record.place_description = place_type
        # place_record.url (not provided)
        # need to consult post data for lat/lng coordinates
        # set to null island first
        lbsn_place_record.geom_center = "POINT(%s %s)" % (0, 0)
        return lbsn_place_record

    @staticmethod
    def clean_tags_from_flickr(tag):
        """Clean special vars not allowed in tags.
        """
        characters_to_replace = ('{', '}')
        for char_check in characters_to_replace:
            tag = tag.replace(char_check, '')
        return tag

    def get_license_number_from_license_name(self, license_name: str):
        """Flickr YFCC100M contains only full string names of licenses
        This function converts names to ids.

        Arguments:
            license_name {str} -- YFCC100M license name
        see:
        https://www.flickr.com/services/api/flickr.photos.licenses.getInfo.html
        """
        lic_number = self.lic_dict.get(license_name)
        return lic_number

    @staticmethod
    def flickr_map_geoaccuracy(flickr_geo_accuracy_level):
        """Flickr Geoaccuracy Levels (16) are mapped to four LBSNstructure levels:
           LBSN PostGeoaccuracy: UNKNOWN = 0; LATLNG = 1; PLACE = 2; CITY = 3;
           COUNTRY = 4
           Fickr: World level is 1, Country is ~3, Region ~6, City ~11,
           Street ~16.

           Flickr Current range is 1-16. Defaults to 16 if not specified.

        Attributes:
        flickr_geo_accuracy_level   Geoaccuracy Level returned from Flickr
        (String, e.g.: "Level12")
        """
        lbsn_geoaccuracy = False
        stripped_level = flickr_geo_accuracy_level.lstrip('Level').strip()
        if stripped_level.isdigit():
            stripped_level = int(stripped_level)
            if stripped_level >= 15:
                lbsn_geoaccuracy = lbsnPost.LATLNG
            elif stripped_level >= 12:
                lbsn_geoaccuracy = lbsnPost.PLACE
            elif stripped_level >= 8:
                lbsn_geoaccuracy = lbsnPost.CITY
            else:
                lbsn_geoaccuracy = lbsnPost.COUNTRY
        else:
            if flickr_geo_accuracy_level == "Street":
                lbsn_geoaccuracy = lbsnPost.LATLNG
            elif flickr_geo_accuracy_level in ("City", "Region"):
                lbsn_geoaccuracy = lbsnPost.CITY
            elif flickr_geo_accuracy_level in ("Country", "World"):
                lbsn_geoaccuracy = lbsnPost.COUNTRY
        return lbsn_geoaccuracy

    def flickr_extract_postlatlng(self, record):
        """Basic routine for extracting lat/lng coordinates from post.
           - checks for consistency and errors
           - in case of any issue, entry is submitted to Null island (0, 0)
        """
        lat_entry = record[13]
        lng_entry = record[12]
        if lat_entry == "" and lng_entry == "":
            l_lat, l_lng = 0, 0
        else:
            try:
                l_lng = Decimal(lng_entry)
                l_lat = Decimal(lat_entry)
            except:
                l_lat, l_lng = 0, 0

        if ((l_lat == 0 and l_lng == 0)
                or l_lat > 90 or l_lat < -90
                or l_lng > 180 or l_lng < -180):
            l_lat, l_lng = 0, 0
            self.send_to_null_island(lat_entry, lng_entry, record[5])
        return FieldMappingYFCC100M.lat_lng_to_wkt(l_lat, l_lng)

    @staticmethod
    def lat_lng_to_wkt(lat, lng):
        """Convert lat lng to WKT (Well-Known-Text)
        """
        point_latlng_string = "POINT(%s %s)" % (lng, lat)
        return point_latlng_string

    def send_to_null_island(self, lat_entry, lng_entry, record_guid):
        """Logs entries with problematic lat/lng's,
           increases Null Island Counter by 1.
        """
        # self.log.debug(f'NULL island: Guid {record_guid} - '
        #               f'Coordinates: {lat_entry}, {lng_entry}')
        self.null_island += 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Collection of classes to import, convert and export data
from and to common lbsn structure

For more info, see [concept](https://gitlab.vgiscience.de/lbsn/concept)
"""

from .lbsntransform_ import LBSNTransform
from .classes.helper_functions import (GeocodeLocations, HelperFunctions,
                                       LBSNRecordDicts, TimeMonitor)
from .classes.load_data import LoadData
from .classes.shared_structure_proto_lbsndb import ProtoLBSNMapping
from .config.config import BaseConfig



# -*- coding: utf-8 -*-

"""
Module for loding data from different sources (CSV, DB, JSON etc.).
"""

import codecs
import csv
import json
import os
import sys
from contextlib import closing
from glob import glob
from pathlib import Path
from typing import Dict, TextIO, List, Union, Any, Generator, Iterator

import ntpath
import requests
from lbsnstructure.lbsnstructure_pb2 import (CompositeKey, Language,
                                             RelationshipKey, lbsnCity,
                                             lbsnCountry, lbsnOrigin,
                                             lbsnPlace, lbsnPost,
                                             lbsnPostReaction,
                                             lbsnRelationship, lbsnUser,
                                             lbsnUserGroup)

from .db_connection import DBConnection
from .helper_functions import GeocodeLocations
from .helper_functions import HelperFunctions as HF


class LoadData():
    """
    Class for loding data from different sources (CSV, DB, JSON etc.).
    """

    def __init__(
            self, importer=None, is_local_input=None,
            startwith_db_rownumber=None,
            skip_until_file=None, cursor_input=None, input_path=None,
            recursive_load=None, local_file_type=None,
            endwith_db_rownumber=None, is_stacked_json=None, csv_delim=None,
            input_lbsn_type=None, geocode_locations=None,
            ignore_input_source_list=None, disable_reactionpost_ref=None,
            map_relations=None, transfer_reactions=None,
            ignore_non_geotagged=None, min_geoaccuracy=None, source_web=None):
        self.is_local_input = is_local_input
        self.start_number = None

        if not self.is_local_input:
            # Start Value, Modify to continue from last processing
            self.continue_number = startwith_db_rownumber
            self.start_number = startwith_db_rownumber
        else:
            if skip_until_file is None:
                self.continue_number = 0
            else:
                self.continue_number = skip_until_file
            self.start_number = 1
        self.source_web = source_web
        self.cursor_input = None
        if self.is_local_input and not self.source_web:
            self.filelist = LoadData._read_local_files(
                input_path=input_path, recursive_load=recursive_load,
                local_file_type=local_file_type,
                skip_until_file=skip_until_file)
            self.cursor_input = cursor_input
        elif self.is_local_input and self.source_web:
            self.filelist = [input_path]
        else:
            self.filelist = None
            # self.cursor_input = LoadData.initialize_connection(
            #    cfg)
        self.finished = False
        self.db_row_number = 0
        self.endwith_db_rownumber = endwith_db_rownumber
        self.is_stacked_json = is_stacked_json
        self.local_file_type = local_file_type
        self.csv_delim = csv_delim
        self.file_format = local_file_type
        self.input_lbsn_type = input_lbsn_type
        self.cursor = None
        self.start_id = None
        #self.transferlimit = cfg.transferlimit
        # Optional Geocoding
        self.geocode_dict = None
        if geocode_locations:
            self.geocode_dict = LoadData.load_geocodes(
                geocode_locations)
        # Optional ignore input sources
        self.ignore_sources_set = None
        if ignore_input_source_list:
            self.ignore_sources_set = LoadData.load_ignore_sources(
                ignore_input_source_list)

        # initialize field mapping structure
        self.import_mapper = importer(
            disable_reactionpost_ref,
            self.geocode_dict,
            map_relations,
            transfer_reactions,
            ignore_non_geotagged,
            self.ignore_sources_set,
            min_geoaccuracy)
        self.finished = False

    def __enter__(self) -> Iterator[Union[
            CompositeKey, Language, RelationshipKey, lbsnCity,
            lbsnCountry, lbsnOrigin, lbsnPlace, lbsnPost,
            lbsnPostReaction, lbsnRelationship, lbsnUser,
            lbsnUserGroup]]:
        """Main pipeline for reading input data

        Combine multiple generators to single pipeline,
        returned for being processed by with-statement
        """
        record_pipeline = self.convert_records(self._process_input(
            self._open_input_files()))
        return record_pipeline
        # return record_pipeline

    def __exit__(self, c_type, value, traceback):
        """Contextmanager exit: nothing to do here"""
        return False

    def _open_input_files(self, count: bool = None):
        """Loops input input filelist and
        returns opened file handles
        """
        if self.cursor_input or self.source_web:
            return None
        # process localfiles
        for file_name in self.filelist:
            self.continue_number += 1
            # HF.log_main_debug(f'\nCurrent file: {ntpath.basename(file_name)}')
            yield open(file_name, 'r', encoding="utf-8", errors='replace')

    def _process_input(self, file_handles: Iterator[TextIO]) -> Iterator[Dict[
            str, Any]]:
        """File parse for CSV or JSON from open file handle

        Output: produces a list of post that can be parsed
        """
        if self.source_web:
            # single web file query
            url = self.filelist[0]
            with closing(requests.get(url, stream=True)) as file_handle:
                record_reader = csv.reader(
                    codecs.iterdecode(
                        file_handle.iter_lines(), 'utf-8'),  # pylint: disable=maybe-no-member
                    delimiter=self.csv_delim,
                    quotechar='"', quoting=csv.QUOTE_NONE)
                for record in record_reader:
                    yield record
        elif self.is_local_input:
            # local file loop
            for file_handle in file_handles:
                if not self.file_format == 'json':
                    while file_handle:
                        record = self.fetch_record_from_file(
                            file_handle)
                        yield record
                else:
                    records = self.fetch_json_data_from_file(
                        file_handle)
                    if records is not None:
                        for record in records:
                            yield record
        else:
            # db query
            while self.cursor:
                record = self.fetch_json_data_from_lbsn(
                    self.cursor, self.continue_number)
                yield record

    def convert_records(self, records: Dict[str, Any]) -> Iterator[Union[
            CompositeKey, Language, RelationshipKey, lbsnCity,
            lbsnCountry, lbsnOrigin, lbsnPlace, lbsnPost,
            lbsnPostReaction, lbsnRelationship, lbsnUser,
            lbsnUserGroup]]:
        """Loops input json or csv records,
        converts to ProtoBuf structure and adds to records_dict

        Returns statistic-counts, modifies (adds results to) import_mapper
        """
        for record in records:
            if self.is_local_input:
                single_record = record
            else:
                self.db_row_number = record[0]
                single_record = record[2]
            # test for empty or malformed records
            if LoadData.skip_empty_or_other(single_record):
                yield None
            if self.local_file_type == 'json' or not self.is_local_input:
                # note: db-records always returned as json
                lbsn_records = self.import_mapper.parse_json_record(
                    single_record, self.input_lbsn_type)
            elif self.local_file_type in ('txt', 'csv'):
                lbsn_records = self.import_mapper.parse_csv_record(
                    single_record)
            else:
                exit(f'Format {self.local_file_type} not supportet.')
            # return record as pipe
            yield lbsn_records

    @staticmethod
    def skip_empty_or_other(single_record):
        """Detect  Rate Limiting Notice or empty records
           so they can be skipped.
        """
        skip = False
        if not single_record or \
            (isinstance(single_record, dict) and
             single_record.get('limit')):
            skip = True
        return skip

    def fetch_json_data_from_lbsn(self, cursor, start_id=0, get_max=None,
                                  number_of_records_to_fetch=10000):
        """Fetches records from Postgres DB

        Keyword arguments:
        cursor -- db-cursor
        start_id -- Offset for querying
        get_max -- optional limit for retrieving records
        number_of_records_to_fetch -- how many records should get fetched
        """
        # if transferlimit is below number_of_records_to_fetch, e.g.  10000,
        # retrieve only necessary volume of records
        if get_max:
            number_of_records_to_fetch = min(
                number_of_records_to_fetch, get_max)
        query_sql = '''
                SELECT in_id,insert_time,data::json FROM public."input"
                WHERE in_id > %s
                ORDER BY in_id ASC LIMIT %s;
                '''
        cursor.execute(query_sql, (start_id, number_of_records_to_fetch))
        records = cursor.fetchall()
        if cursor.rowcount == 0:
            return None
        # update last returned db_row_number
        self.continue_number = records[-1][0]
        if not self.start_number:
            # first returned db_row_number
            self.start_number = records[0][0]
        for record in records:
            return record

    def fetch_record_from_file(self, file_handle):
        """Fetches CSV or JSON data (including stacked json) from file"""

        if self.file_format in ['txt', 'csv']:
            record_reader = self.fetch_csv_data_from_file(
                file_handle)
        else:
            exit(f'Format {self.file_format} not supported.')
        # return record pipeline
        for record in record_reader:
            return record

    def fetch_json_data_from_file(self, file_handle, start_file_id=0):
        """Read json entries from file.

        Typical form is [{json1},{json2}], if is_stacked_json is True:
        will process stacked jsons in the form of {json1}{json2}
        """
        records = []
        # Stacked JSON is a simple file with many concatenated jsons, e.g.
        # {json1}{json2} etc.
        if self.is_stacked_json:
            try:
                for obj in HF.decode_stacked(file_handle.read()):
                    records.append(obj)
            except json.decoder.JSONDecodeError:
                pass
        else:
            # normal json nesting, e.g.  {{record1},{record2}}
            try:
                records = json.load(file_handle)
            except json.decoder.JSONDecodeError:
                pass
        if records:
            return records
        return None

    def fetch_csv_data_from_file(self, file_handle):
        """Read csv entries from file (either *.txt or *.csv).

        The actual CSV formatting is not setable in config yet.
        There are many specifics, e.g.
        #QUOTE_NONE is used here because media saved from Flickr
        does not contain any quotes ""
        """
        if self.csv_delim is None:
            self.csv_delim = ','
        #records = []
        record_reader = csv.reader(file_handle, delimiter=self.csv_delim,
                                   quotechar='"', quoting=csv.QUOTE_NONE)
        return record_reader
        #   next(reader, None)  # skip headerline
        #   records = list(reader)
        #   if not records:
        #       return None
        #   return records

    @staticmethod
    def initialize_connection(
            dbuser_output, dbserveraddress_output,
            dbname_output, dbpassword_output, dbserverport_output,
            readonly: bool = True):
        """Establishes connection to DB (Postgres)"""

        if dbuser_output:
            connection = DBConnection(
                serveraddress=dbserveraddress_output,
                dbname=dbname_output,
                user=dbuser_output,
                password=dbpassword_output,
                port=dbserverport_output,
                readonly=readonly)
            conn, cursor = connection.connect()
        else:
            conn = None
            cursor = None
        return conn, cursor

    @staticmethod
    def _read_local_files(input_path=None, recursive_load=None,
                          local_file_type=None, skip_until_file=None):
        """Read Local Files according to config parameters and
        returns list of file-paths
        """
        if recursive_load:
            excludefolderlist = ["01_DataSetHistory",
                                 "02_UserData", "03_ClippedData", "04_MapVis"]
            excludestartswithfile = ["log", "settings", "GridCoordinates"]
            loc_filelist = \
                LoadData.scan_rec(input_path,
                                  file_format=local_file_type,
                                  excludefolderlist=excludefolderlist,
                                  excludestartswithfile=excludestartswithfile)
        else:
            loc_filelist_gen = input_path.glob(f'*.{local_file_type}')
            loc_filelist = []
            for file_path in loc_filelist_gen:
                loc_filelist.append(file_path)
        if skip_until_file:
            print("Implement skip until file")
        input_count = (len(loc_filelist))
        if input_count == 0:
            sys.exit("No location files found.")
        else:
            return loc_filelist

    @staticmethod
    def load_geocodes(geo_config):
        """Loads coordinates-string tuples for geocoding
        text locations (Optional)
        """
        locationsgeocode_dict = GeocodeLocations()
        locationsgeocode_dict.load_geocodelist(geo_config)
        geocode_dict = locationsgeocode_dict.geocode_dict
        return geocode_dict

    @staticmethod
    def load_ignore_sources(list_source=None):
        """Loads list of source types to be ignored"""
        if list_source is None:
            return
        ignore_source_list = set()
        with open(list_source, newline='', encoding='utf8') as file_handle:
            for ignore_source in file_handle:
                ignore_source_list.add(ignore_source.strip())
        return ignore_source_list

    @staticmethod
    def scan_rec(root, subdirlimit=2, file_format="csv",
                 excludefolderlist=None, excludestartswithfile=None):
        """Recursively scan subdir for datafiles"""
        rval = []
        if excludefolderlist is None:
            excludefolderlist = []
        if excludestartswithfile is None:
            excludestartswithfile = []

        def do_scan(start_dir, output, depth=0):
            for file_handle in os.listdir(start_dir):
                file_handle_path = os.path.join(start_dir, file_handle)
                if os.path.isdir(file_handle_path):
                    if depth < subdirlimit:
                        efound = False
                        # check for excludefolders:
                        for entry in excludefolderlist:
                            if entry in file_handle_path:
                                efound = True
                                break
                        if efound is False:
                            do_scan(file_handle_path, output, depth+1)
                else:
                    if file_handle_path.endswith(file_format):
                        efound = False
                        for entry in excludestartswithfile:
                            if ntpath.basename(
                                    file_handle_path).startswith(entry):
                                efound = True
                                break
                        if efound is False:
                            output.append(file_handle_path)
        do_scan(root, rval, 0)
        return rval
